from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import cast, List


T = TypeVar("T", bound="ReportExtras")


@_attrs_define
class ReportExtras:
    """
    Attributes:
        report_uuid (str):
        network_name (str):
        cf_violations (int):
        ap_violations (int):
        status (str):
        errors_count (int):
        errors_lines (List[str]):
    """

    report_uuid: str
    network_name: str
    cf_violations: int
    ap_violations: int
    status: str
    errors_count: int
    errors_lines: List[str]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        report_uuid = self.report_uuid
        network_name = self.network_name
        cf_violations = self.cf_violations
        ap_violations = self.ap_violations
        status = self.status
        errors_count = self.errors_count
        errors_lines = self.errors_lines

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "report_uuid": report_uuid,
                "network_name": network_name,
                "cf_violations": cf_violations,
                "ap_violations": ap_violations,
                "status": status,
                "errors_count": errors_count,
                "errors_lines": errors_lines,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        report_uuid = d.pop("report_uuid")

        network_name = d.pop("network_name")

        cf_violations = d.pop("cf_violations")

        ap_violations = d.pop("ap_violations")

        status = d.pop("status")

        errors_count = d.pop("errors_count")

        errors_lines = cast(List[str], d.pop("errors_lines"))

        report_extras = cls(
            report_uuid=report_uuid,
            network_name=network_name,
            cf_violations=cf_violations,
            ap_violations=ap_violations,
            status=status,
            errors_count=errors_count,
            errors_lines=errors_lines,
        )

        report_extras.additional_properties = d
        return report_extras

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



# Invariant client authentication library


from dataclasses import dataclass
import enum
import ssl
from typing import Optional
from invariant_client import pysdk

from invariant_client.pysdk import AccessCredential, RemoteError

from invariant_client.bindings.invariant_login_client import models
from invariant_client.bindings.invariant_login_client.client import AuthenticatedClient as LoginAuthenticatedClient, Client as LoginClient
from invariant_client.bindings.invariant_login_client.api.login.get_version_api_v1_login_version_get import sync_detailed as get_version_api_v1_login_version_get


class VersionClient:
    """Connect to the server and get the current version."""

    client_base_url: str
    client_kwargs: dict

    def __init__(self,
            base_url: Optional[str] = None,
            verify_ssl: Optional[str | bool | ssl.SSLContext] = None,
            **kwargs,
        ):
        self.client_base_url = base_url or pysdk.DOMAIN_NAME
        self.client_kwargs = {}
        self.client_kwargs.update(kwargs)
        self.client_kwargs['verify_ssl'] = verify_ssl

    def get_version(self) -> str:
        """Connect to invariant.tech and get the current version."""
        client = LoginClient(self.client_base_url, **self.client_kwargs)
        response = get_version_api_v1_login_version_get(client=client)
        response = response.parsed
        if not response:
            raise RemoteError(f"Unable to connect to {self.client_base_url}")
        if not isinstance(response, models.GetVersionResponse):
            raise RemoteError(response)
        return response.version



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Literal
from typing import cast
from typing import Dict
from typing import cast, List

if TYPE_CHECKING:
    from ..models.github_branch import GithubBranch
    from ..models.github_repository import GithubRepository


T = TypeVar("T", bound="GithubRepositoryData")


@_attrs_define
class GithubRepositoryData:
    """
    Attributes:
        integration_uuid (str):
        type (Literal['github']):
        stub (bool):
        url (str):
        github_repo (GithubRepository):
        branches (List['GithubBranch']):
        sot_branch (List[str]):
        sync_branches (List[str]):
    """

    integration_uuid: str
    type: Literal["github"]
    stub: bool
    url: str
    github_repo: "GithubRepository"
    branches: List["GithubBranch"]
    sot_branch: List[str]
    sync_branches: List[str]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        integration_uuid = self.integration_uuid
        type = self.type
        stub = self.stub
        url = self.url
        github_repo = self.github_repo.to_dict()

        branches = []
        for branches_item_data in self.branches:
            branches_item = branches_item_data.to_dict()

            branches.append(branches_item)

        sot_branch = self.sot_branch

        sync_branches = self.sync_branches

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "integration_uuid": integration_uuid,
                "type": type,
                "stub": stub,
                "url": url,
                "github_repo": github_repo,
                "branches": branches,
                "sot_branch": sot_branch,
                "sync_branches": sync_branches,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.github_branch import GithubBranch
        from ..models.github_repository import GithubRepository

        d = src_dict.copy()
        integration_uuid = d.pop("integration_uuid")

        type = d.pop("type")

        stub = d.pop("stub")

        url = d.pop("url")

        github_repo = GithubRepository.from_dict(d.pop("github_repo"))

        branches = []
        _branches = d.pop("branches")
        for branches_item_data in _branches:
            branches_item = GithubBranch.from_dict(branches_item_data)

            branches.append(branches_item)

        sot_branch = cast(List[str], d.pop("sot_branch"))

        sync_branches = cast(List[str], d.pop("sync_branches"))

        github_repository_data = cls(
            integration_uuid=integration_uuid,
            type=type,
            stub=stub,
            url=url,
            github_repo=github_repo,
            branches=branches,
            sot_branch=sot_branch,
            sync_branches=sync_branches,
        )

        github_repository_data.additional_properties = d
        return github_repository_data

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Dict
from typing import List

if TYPE_CHECKING:
    from ..models.notification_group import NotificationGroup


T = TypeVar("T", bound="ListNotificationGroupsResponse")


@_attrs_define
class ListNotificationGroupsResponse:
    """List of NotificationGroups

    Attributes:
        notification_groups (List['NotificationGroup']):
    """

    notification_groups: List["NotificationGroup"]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        notification_groups = []
        for notification_groups_item_data in self.notification_groups:
            notification_groups_item = notification_groups_item_data.to_dict()

            notification_groups.append(notification_groups_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "notification_groups": notification_groups,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.notification_group import NotificationGroup

        d = src_dict.copy()
        notification_groups = []
        _notification_groups = d.pop("notification_groups")
        for notification_groups_item_data in _notification_groups:
            notification_groups_item = NotificationGroup.from_dict(
                notification_groups_item_data
            )

            notification_groups.append(notification_groups_item)

        list_notification_groups_response = cls(
            notification_groups=notification_groups,
        )

        list_notification_groups_response.additional_properties = d
        return list_notification_groups_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import cast, Union
from typing import cast, List


T = TypeVar("T", bound="CreateNotificationGroupRequest")


@_attrs_define
class CreateNotificationGroupRequest:
    """
    Attributes:
        name (str):
        comment (str):
        email_targets (Union[List[str], None]):
        network_subscriptions (Union[List[str], None]):
    """

    name: str
    comment: str
    email_targets: Union[List[str], None]
    network_subscriptions: Union[List[str], None]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        comment = self.comment
        email_targets: Union[List[str], None]

        if isinstance(self.email_targets, list):
            email_targets = self.email_targets

        else:
            email_targets = self.email_targets

        network_subscriptions: Union[List[str], None]

        if isinstance(self.network_subscriptions, list):
            network_subscriptions = self.network_subscriptions

        else:
            network_subscriptions = self.network_subscriptions

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "comment": comment,
                "email_targets": email_targets,
                "network_subscriptions": network_subscriptions,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        comment = d.pop("comment")

        def _parse_email_targets(data: object) -> Union[List[str], None]:
            if data is None:
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                email_targets_type_0 = cast(List[str], data)

                return email_targets_type_0
            except:  # noqa: E722
                pass
            return cast(Union[List[str], None], data)

        email_targets = _parse_email_targets(d.pop("email_targets"))

        def _parse_network_subscriptions(data: object) -> Union[List[str], None]:
            if data is None:
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                network_subscriptions_type_0 = cast(List[str], data)

                return network_subscriptions_type_0
            except:  # noqa: E722
                pass
            return cast(Union[List[str], None], data)

        network_subscriptions = _parse_network_subscriptions(
            d.pop("network_subscriptions")
        )

        create_notification_group_request = cls(
            name=name,
            comment=comment,
            email_targets=email_targets,
            network_subscriptions=network_subscriptions,
        )

        create_notification_group_request.additional_properties = d
        return create_notification_group_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Type, TypeVar

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from dateutil.parser import isoparse
import datetime


T = TypeVar("T", bound="ReportTask")


@_attrs_define
class ReportTask:
    """
    Attributes:
        uuid (str):
        organization_uuid (str):
        created_at (datetime.datetime):
        urn (str):
        type (str):
        initiator_urn (str):
        worker_pod (str):
        was_killed (bool):
    """

    uuid: str
    organization_uuid: str
    created_at: datetime.datetime
    urn: str
    type: str
    initiator_urn: str
    worker_pod: str
    was_killed: bool
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        uuid = self.uuid
        organization_uuid = self.organization_uuid
        created_at = self.created_at.isoformat()

        urn = self.urn
        type = self.type
        initiator_urn = self.initiator_urn
        worker_pod = self.worker_pod
        was_killed = self.was_killed

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "uuid": uuid,
                "organization_uuid": organization_uuid,
                "created_at": created_at,
                "urn": urn,
                "type": type,
                "initiator_urn": initiator_urn,
                "worker_pod": worker_pod,
                "was_killed": was_killed,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        uuid = d.pop("uuid")

        organization_uuid = d.pop("organization_uuid")

        created_at = isoparse(d.pop("created_at"))

        urn = d.pop("urn")

        type = d.pop("type")

        initiator_urn = d.pop("initiator_urn")

        worker_pod = d.pop("worker_pod")

        was_killed = d.pop("was_killed")

        report_task = cls(
            uuid=uuid,
            organization_uuid=organization_uuid,
            created_at=created_at,
            urn=urn,
            type=type,
            initiator_urn=initiator_urn,
            worker_pod=worker_pod,
            was_killed=was_killed,
        )

        report_task.additional_properties = d
        return report_task

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import enum
import json
import os
import pydoc
import sys
from typing import Any, Iterable, Mapping
from attr import asdict
import pandas
from rich import print_json
from tabulate import tabulate
from invariant_client.bindings.invariant_instance_client.models.get_report_summary_response import GetReportSummaryResponse


class OutputFormat(enum.Enum):
    TABULATE = enum.auto()
    JSON = enum.auto()
    TSV = enum.auto()
    FAST_JSON = enum.auto()
    CONDENSED = enum.auto()
    # etc


def snapshot_status(response: GetReportSummaryResponse):
    """Display a message indicating the overall status of the analysis user task."""
    status = response.status.to_dict()
    status = status.get('state')
    if status == 'COMPLETE':
        pass
    elif status == 'FAILED':
        print("Error: Snapshot could not be evaluated.")
    elif status == 'INCOMPLETE':
        print("Error: Snapshot evaluation did not finish.")


# AP_KEYS = [
#     'critical_flows_ok', 'critical_flows_violations', 'critical_flows_violations_unenforced', 'critical_flows_skipped', 'critical_flows_details', 'critical_flows_logs',
#     'policy_ok', 'policy_violations', 'policy_violations_unenforced', 'policy_skipped', 'policy_details', 'policy_logs']


class CondensedOutcomes(str, enum.Enum):
    OK = "All rules passed"
    VIOLATIONS = "Rule violations found"
    NO_MODEL = "Could not model network"
    NO_RULES = "No rules provided"
    SKIPPED = "Some rules skipped"
    ERROR = "Processing failed"

    @classmethod
    def from_GetReportSummaryResponse(cls, response: GetReportSummaryResponse):
        """Map this response onto an outcome."""
        # Total disaster
        if response.status['state'] == 'FAILED':
            return cls.ERROR

        # No model could be created at all
        summary = response.summary.to_dict()
        if summary.get('nodes', None) is None:
            return cls.NO_MODEL

        # No rules were provided
        if summary.get('policy_violations', None) is None:
            return cls.NO_RULES

        # Violations found
        if summary.get('policy_violations', 0) + summary.get('critical_flows_violations', 0) > 0:
            return cls.VIOLATIONS

        # OK but some invalid rules were found
        if summary.get('policy_skipped', 0) + summary.get('critical_flows_skipped', 0) > 0:
            return cls.SKIPPED

        # OK
        if summary.get('policy_ok', 0) + summary.get('critical_flows_ok', 0) > 0:
            return cls.OK

        # if sum(summary.get(key, 0) for key in AP_KEYS) == 0:
        #     return cls.NO_RULES
        # Fall back to no-rules
        return cls.NO_RULES


def snapshot_condensed_status(response: GetReportSummaryResponse):
    """Display a single message indicating the overall snapshot outcome."""
    outcome = CondensedOutcomes.from_GetReportSummaryResponse(response)
    print(f"outcome: {outcome.value}")


def snapshot_summary_table(response: GetReportSummaryResponse, format: OutputFormat):
    """Display a table containing row counts for all emitted reports."""
    if format == OutputFormat.JSON:
        print_json(data=response.to_dict())
    elif format == OutputFormat.FAST_JSON:
        print(json.dumps(response.to_dict()))
    else:
        print_frame(pandas.DataFrame(response.summary.to_dict().items(), columns=['File', 'RowCount']), format)


def snapshot_halted(response: GetReportSummaryResponse):
    """Describe each halted step."""
    status = response.status.to_dict()
    halted = []
    def visit_step(step: dict, prefix: list):
        prefix = prefix + [step['name']] if prefix else [step['name']]
        if step['state'] != 'COMPLETE':
            halted.append((prefix, step))
        for child_step in step['steps']:
            visit_step(child_step, prefix)
    visit_step(status, None)
    if len(halted):
        print("\nThe following steps were not completed:")
        for prefix, step in halted:
            print(f"    {' > '.join(prefix)}")
            if step['state'] != 'FAILED':
                print(f"\n        {step['state']}")


def snapshot_errors(errors: pandas.DataFrame, format: OutputFormat):
    """Describe each error."""
    # Group errors by label, then repeat the header every 10th member
    for label, error_group in errors.groupby(by=['label']):
        label = label[0]
        for i, error in enumerate(error_group.to_dict(orient='records')):
            prefix = ""
            if not i % 10:
                prefix = f"In {label}{' (continued)' if i > 0 else ''}:\n\n"
            data = json.loads(error['detail'])
            if data['type'] == 'urn:invariant:errors:child_step_failed':
                continue
            if data['type'] == 'urn:invariant:errors:internal_exception':
                print(f"\nInternal error in {label}:\n    {data['detail']}", file=sys.stderr)
                continue

            print(f"\n{prefix}    {data['title']}\n    {data['detail']}", file=sys.stderr)
    print('', file=sys.stderr)

def print_frame(data: Mapping[str, Iterable[Any]] | Iterable[Iterable[Any]], format: OutputFormat):
    if format == OutputFormat.TSV:
        print(tabulate(data, headers='keys', tablefmt='tsv'))
    elif format == OutputFormat.TABULATE or format == OutputFormat.CONDENSED:
        os.environ["LESS"] = "-SXFRs"
        pydoc.pager(tabulate(data, headers='keys', tablefmt='psql'))
    else:
        raise ValueError(f"Unacceptable format: {format}")



from typing import Any, Dict, Type, TypeVar, TYPE_CHECKING

from typing import List


from attrs import define as _attrs_define
from attrs import field as _attrs_field


from typing import Dict
from typing import List

if TYPE_CHECKING:
    from ..models.api_token import APIToken


T = TypeVar("T", bound="APITokenResponse")


@_attrs_define
class APITokenResponse:
    """List of APITokens

    Attributes:
        api_tokens (List['APIToken']):
    """

    api_tokens: List["APIToken"]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        api_tokens = []
        for api_tokens_item_data in self.api_tokens:
            api_tokens_item = api_tokens_item_data.to_dict()

            api_tokens.append(api_tokens_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "api_tokens": api_tokens,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.api_token import APIToken

        d = src_dict.copy()
        api_tokens = []
        _api_tokens = d.pop("api_tokens")
        for api_tokens_item_data in _api_tokens:
            api_tokens_item = APIToken.from_dict(api_tokens_item_data)

            api_tokens.append(api_tokens_item)

        api_token_response = cls(
            api_tokens=api_tokens,
        )

        api_token_response.additional_properties = d
        return api_token_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Dict, Optional, Union

import httpx

from ...client import AuthenticatedClient, Client
from ...types import Response, UNSET
from ... import errors

from typing import Union
from ...models.challenge_response import ChallengeResponse
from typing import Dict
from ...models.list_reports_response import ListReportsResponse
from ...models.validation_error_response import ValidationErrorResponse
from ...models.base_error_response import BaseErrorResponse
from ...types import UNSET, Unset
from typing import Union


def _get_kwargs(
    organization_name: str,
    *,
    filter_session: Union[None, Unset, int] = 0,
    filter_net: Union[None, Unset, str] = UNSET,
    limit: Union[None, Unset, int] = 0,
) -> Dict[str, Any]:
    params: Dict[str, Any] = {}
    json_filter_session: Union[None, Unset, int]
    if isinstance(filter_session, Unset):
        json_filter_session = UNSET

    else:
        json_filter_session = filter_session

    params["filter_session"] = json_filter_session

    json_filter_net: Union[None, Unset, str]
    if isinstance(filter_net, Unset):
        json_filter_net = UNSET

    else:
        json_filter_net = filter_net

    params["filter_net"] = json_filter_net

    json_limit: Union[None, Unset, int]
    if isinstance(limit, Unset):
        json_limit = UNSET

    else:
        json_limit = limit

    params["limit"] = json_limit

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    return {
        "method": "get",
        "url": "/{organization_name}/api/v1/reports/".format(
            organization_name=organization_name,
        ),
        "params": params,
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ListReportsResponse,
        ValidationErrorResponse,
    ]
]:
    if response.status_code == HTTPStatus.OK:
        response_200 = ListReportsResponse.from_dict(response.json())

        return response_200
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = BaseErrorResponse.from_dict(response.json())

        return response_404
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ListReportsResponse,
        ValidationErrorResponse,
    ]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    filter_session: Union[None, Unset, int] = 0,
    filter_net: Union[None, Unset, str] = UNSET,
    limit: Union[None, Unset, int] = 0,
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ListReportsResponse,
        ValidationErrorResponse,
    ]
]:
    """Lists all reports for an organization. Note that filter_session is not compatible with other
    filters.

    Args:
        organization_name (str):
        filter_session (Union[None, Unset, int]):
        filter_net (Union[None, Unset, str]):
        limit (Union[None, Unset, int]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[BaseErrorResponse, ChallengeResponse, ListReportsResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        filter_session=filter_session,
        filter_net=filter_net,
        limit=limit,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    filter_session: Union[None, Unset, int] = 0,
    filter_net: Union[None, Unset, str] = UNSET,
    limit: Union[None, Unset, int] = 0,
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ListReportsResponse,
        ValidationErrorResponse,
    ]
]:
    """Lists all reports for an organization. Note that filter_session is not compatible with other
    filters.

    Args:
        organization_name (str):
        filter_session (Union[None, Unset, int]):
        filter_net (Union[None, Unset, str]):
        limit (Union[None, Unset, int]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[BaseErrorResponse, ChallengeResponse, ListReportsResponse, ValidationErrorResponse]
    """

    return sync_detailed(
        organization_name=organization_name,
        client=client,
        filter_session=filter_session,
        filter_net=filter_net,
        limit=limit,
    ).parsed


async def asyncio_detailed(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    filter_session: Union[None, Unset, int] = 0,
    filter_net: Union[None, Unset, str] = UNSET,
    limit: Union[None, Unset, int] = 0,
) -> Response[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ListReportsResponse,
        ValidationErrorResponse,
    ]
]:
    """Lists all reports for an organization. Note that filter_session is not compatible with other
    filters.

    Args:
        organization_name (str):
        filter_session (Union[None, Unset, int]):
        filter_net (Union[None, Unset, str]):
        limit (Union[None, Unset, int]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[BaseErrorResponse, ChallengeResponse, ListReportsResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        organization_name=organization_name,
        filter_session=filter_session,
        filter_net=filter_net,
        limit=limit,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    organization_name: str,
    *,
    client: AuthenticatedClient,
    filter_session: Union[None, Unset, int] = 0,
    filter_net: Union[None, Unset, str] = UNSET,
    limit: Union[None, Unset, int] = 0,
) -> Optional[
    Union[
        BaseErrorResponse,
        ChallengeResponse,
        ListReportsResponse,
        ValidationErrorResponse,
    ]
]:
    """Lists all reports for an organization. Note that filter_session is not compatible with other
    filters.

    Args:
        organization_name (str):
        filter_session (Union[None, Unset, int]):
        filter_net (Union[None, Unset, str]):
        limit (Union[None, Unset, int]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[BaseErrorResponse, ChallengeResponse, ListReportsResponse, ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            organization_name=organization_name,
            client=client,
            filter_session=filter_session,
            filter_net=filter_net,
            limit=limit,
        )
    ).parsed



from http import HTTPStatus
from typing import Any, Dict, Optional, Union, cast

import httpx

from ...client import AuthenticatedClient, Client
from ...types import Response
from ... import errors

from ...models.login_request import LoginRequest
from typing import cast
from typing import Dict
from ...models.base_error_response import BaseErrorResponse
from ...models.challenge_response import ChallengeResponse
from ...models.validation_error_response import ValidationErrorResponse


def _get_kwargs(
    *,
    json_body: LoginRequest,
) -> Dict[str, Any]:
    json_json_body = json_body.to_dict()

    return {
        "method": "post",
        "url": "/api/v1/login/",
        "json": json_json_body,
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    if response.status_code == HTTPStatus.NO_CONTENT:
        response_204 = cast(Any, None)
        return response_204
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ValidationErrorResponse.from_dict(response.json())

        return response_422
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = ChallengeResponse.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = BaseErrorResponse.from_dict(response.json())

        return response_404
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient,
    json_body: LoginRequest,
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Authenticate and respond to authentication challenges

     To authenticate, send a request with basic_auth provided.

    If additional steps are neded by the user, this API will send a ChallengeResponse (HTTP 401),
    which includes a challenge type and a Login Session Token. Use this API to respond to the challenge
    and send the given token.

    On successful authentication an HTTPOnly refresh token cookie will be issued through a set-cookie
    header.

    Note: this is not the only authentication entrypoint. It is also possible to authenticate from a
    single-use email invitation link using the /check_invite method.

    ## Using the cookie

    The HTTPOnly refresh token cookie can be used to:

    -   Get a list of available instances: /get_instances.
    -   Get an instance access token: /:org/api/v1/refresh.
    -   Refresh the refresh token before it expires: /refresh.

    ## Basic authentication

        Key: basic_auth
        Authentication: none

    Provide an email and password value. If the password is incorrect a ChallengeResponse will be issued
    with
    challenge=basic_auth.

    ## Email verification

        Key: email_valid
        Authentication: Login Session Token

    Provide a secret PIN value from the verification email. The PIN must be non-expired and is consumed
    on use.
    /resend_validation_pin can be used to re-send a verification email.

    ## Password reset

        Key: reset_request
        Authentication: none

    Provide an email to send a password reset email to it. A ChallengeResponse will be issued with
    challenge=reset_pin .

    ## Password reset check PIN

        Key: reset_pin_request
        Authentication: Login Session Token

    Provide a secret PIN value from the password reset email. The PIN must be non-expired. Checking the
    PIN
    does not consume it, it will be re-checked and consumed when the user provides a new password. A
    ChallengeResponse will be issued with challenge=new_password .

    ## New password

        Key: new_password
        Authentication: Login Session Token

    Provide a new password and a secret PIN value from the password reset email OR an email invitation
    token (which is a JWT
    containing a secret PIN). The PIN must be non-expired and will be consumed on use.

    Args:
        json_body (LoginRequest): Request to determine login method (Basic, SAML, etc).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        json_body=json_body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: AuthenticatedClient,
    json_body: LoginRequest,
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Authenticate and respond to authentication challenges

     To authenticate, send a request with basic_auth provided.

    If additional steps are neded by the user, this API will send a ChallengeResponse (HTTP 401),
    which includes a challenge type and a Login Session Token. Use this API to respond to the challenge
    and send the given token.

    On successful authentication an HTTPOnly refresh token cookie will be issued through a set-cookie
    header.

    Note: this is not the only authentication entrypoint. It is also possible to authenticate from a
    single-use email invitation link using the /check_invite method.

    ## Using the cookie

    The HTTPOnly refresh token cookie can be used to:

    -   Get a list of available instances: /get_instances.
    -   Get an instance access token: /:org/api/v1/refresh.
    -   Refresh the refresh token before it expires: /refresh.

    ## Basic authentication

        Key: basic_auth
        Authentication: none

    Provide an email and password value. If the password is incorrect a ChallengeResponse will be issued
    with
    challenge=basic_auth.

    ## Email verification

        Key: email_valid
        Authentication: Login Session Token

    Provide a secret PIN value from the verification email. The PIN must be non-expired and is consumed
    on use.
    /resend_validation_pin can be used to re-send a verification email.

    ## Password reset

        Key: reset_request
        Authentication: none

    Provide an email to send a password reset email to it. A ChallengeResponse will be issued with
    challenge=reset_pin .

    ## Password reset check PIN

        Key: reset_pin_request
        Authentication: Login Session Token

    Provide a secret PIN value from the password reset email. The PIN must be non-expired. Checking the
    PIN
    does not consume it, it will be re-checked and consumed when the user provides a new password. A
    ChallengeResponse will be issued with challenge=new_password .

    ## New password

        Key: new_password
        Authentication: Login Session Token

    Provide a new password and a secret PIN value from the password reset email OR an email invitation
    token (which is a JWT
    containing a secret PIN). The PIN must be non-expired and will be consumed on use.

    Args:
        json_body (LoginRequest): Request to determine login method (Basic, SAML, etc).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
    """

    return sync_detailed(
        client=client,
        json_body=json_body,
    ).parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient,
    json_body: LoginRequest,
) -> Response[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Authenticate and respond to authentication challenges

     To authenticate, send a request with basic_auth provided.

    If additional steps are neded by the user, this API will send a ChallengeResponse (HTTP 401),
    which includes a challenge type and a Login Session Token. Use this API to respond to the challenge
    and send the given token.

    On successful authentication an HTTPOnly refresh token cookie will be issued through a set-cookie
    header.

    Note: this is not the only authentication entrypoint. It is also possible to authenticate from a
    single-use email invitation link using the /check_invite method.

    ## Using the cookie

    The HTTPOnly refresh token cookie can be used to:

    -   Get a list of available instances: /get_instances.
    -   Get an instance access token: /:org/api/v1/refresh.
    -   Refresh the refresh token before it expires: /refresh.

    ## Basic authentication

        Key: basic_auth
        Authentication: none

    Provide an email and password value. If the password is incorrect a ChallengeResponse will be issued
    with
    challenge=basic_auth.

    ## Email verification

        Key: email_valid
        Authentication: Login Session Token

    Provide a secret PIN value from the verification email. The PIN must be non-expired and is consumed
    on use.
    /resend_validation_pin can be used to re-send a verification email.

    ## Password reset

        Key: reset_request
        Authentication: none

    Provide an email to send a password reset email to it. A ChallengeResponse will be issued with
    challenge=reset_pin .

    ## Password reset check PIN

        Key: reset_pin_request
        Authentication: Login Session Token

    Provide a secret PIN value from the password reset email. The PIN must be non-expired. Checking the
    PIN
    does not consume it, it will be re-checked and consumed when the user provides a new password. A
    ChallengeResponse will be issued with challenge=new_password .

    ## New password

        Key: new_password
        Authentication: Login Session Token

    Provide a new password and a secret PIN value from the password reset email OR an email invitation
    token (which is a JWT
    containing a secret PIN). The PIN must be non-expired and will be consumed on use.

    Args:
        json_body (LoginRequest): Request to determine login method (Basic, SAML, etc).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]]
    """

    kwargs = _get_kwargs(
        json_body=json_body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: AuthenticatedClient,
    json_body: LoginRequest,
) -> Optional[
    Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
]:
    """Authenticate and respond to authentication challenges

     To authenticate, send a request with basic_auth provided.

    If additional steps are neded by the user, this API will send a ChallengeResponse (HTTP 401),
    which includes a challenge type and a Login Session Token. Use this API to respond to the challenge
    and send the given token.

    On successful authentication an HTTPOnly refresh token cookie will be issued through a set-cookie
    header.

    Note: this is not the only authentication entrypoint. It is also possible to authenticate from a
    single-use email invitation link using the /check_invite method.

    ## Using the cookie

    The HTTPOnly refresh token cookie can be used to:

    -   Get a list of available instances: /get_instances.
    -   Get an instance access token: /:org/api/v1/refresh.
    -   Refresh the refresh token before it expires: /refresh.

    ## Basic authentication

        Key: basic_auth
        Authentication: none

    Provide an email and password value. If the password is incorrect a ChallengeResponse will be issued
    with
    challenge=basic_auth.

    ## Email verification

        Key: email_valid
        Authentication: Login Session Token

    Provide a secret PIN value from the verification email. The PIN must be non-expired and is consumed
    on use.
    /resend_validation_pin can be used to re-send a verification email.

    ## Password reset

        Key: reset_request
        Authentication: none

    Provide an email to send a password reset email to it. A ChallengeResponse will be issued with
    challenge=reset_pin .

    ## Password reset check PIN

        Key: reset_pin_request
        Authentication: Login Session Token

    Provide a secret PIN value from the password reset email. The PIN must be non-expired. Checking the
    PIN
    does not consume it, it will be re-checked and consumed when the user provides a new password. A
    ChallengeResponse will be issued with challenge=new_password .

    ## New password

        Key: new_password
        Authentication: Login Session Token

    Provide a new password and a secret PIN value from the password reset email OR an email invitation
    token (which is a JWT
    containing a secret PIN). The PIN must be non-expired and will be consumed on use.

    Args:
        json_body (LoginRequest): Request to determine login method (Basic, SAML, etc).

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Any, BaseErrorResponse, ChallengeResponse, ValidationErrorResponse]
    """

    return (
        await asyncio_detailed(
            client=client,
            json_body=json_body,
        )
    ).parsed
